/*
 *======================================================================
 *
 *Copyright 1992 Sun Microsystems, Inc.
 *
 *The Interface Definition Language Compiler Front End (CFE) is made
 *available for use provided that this legend is included on all media and
 *documentation and as a part of the software program in whole or part.
 *Users may copy and extend functionality (but may not remove
 *functionality) of the Interface Definition Language CFE without charge,
 *but are not authorized to license or distribute it to anyone else except
 *as part of a product or program developed by the user or with the express
 *written consent of Sun Microsystems, Inc.
 *
 *The names of Sun Microsystems, Inc. and any of its subsidiaries may not
 *be used in advertising or publicity pertaining to distribution of
 *Interface Definition Language CFE as permitted herein.
 *
 *The Interface Definition Language CFE may not be exported outside the
 *United States without first obtaining the appropriate government
 *approvals.
 *
 *INTERFACE DEFINITION LANGUAGE CFE IS PROVIDED AS IS WITH NO WARRANTIES
 *OF ANY KIND INCLUDING THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND
 *FITNESS FOR A PARTICULAR PURPOSE, NONINFRINGEMENT, OR ARISING FROM A
 *COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 *Interface Definition Language CFE is provided with no support and
 *without any obligation on the part of Sun Microsystems, Inc. or any of
 *its subsidiaries or affiliates to assist in its use, correction,
 *modification or enhancement.
 *
 *SUN MICROSYSTEMS, INC. OR ANY OF ITS SUBSIDIARIES OR AFFILIATES SHALL
 *HAVE NO LIABILITY WITH RESPECT TO THE INFRINGEMENT OF COPYRIGHTS, TRADE
 *SECRETS OR ANY PATENTS BY INTERFACE DEFINITION LANGUAGE CFE OR ANY PART
 *THEREOF.
 *
 *IN NO EVENT WILL SUN MICROSYSTEMS, INC. OR ANY OF ITS SUBSIDIARIES OR
 *AFFILIATES BE LIABLE FOR ANY LOST REVENUE OR PROFITS OR OTHER SPECIAL,
 *INDIRECT AND CONSEQUENTIAL DAMAGES, EVEN IF SUN HAS BEEN ADVISED OF THE
 *POSSIBILITY OF SUCH DAMAGES.
 *
 *SunSoft, Inc.  
 *2550 Garcia Avenue 
 *Mountain View, California  94043
 *
 *
 *
 *
 *======================================================================
 */

#ifndef _AST_PREDEFINED_TYPE_AST_PREDEFINED_TYPE_HH
#define _AST_PREDEFINED_TYPE_AST_PREDEFINED_TYPE_HH

#pragma ident "%@(#)AST_PredefinedType.h	1.37% %92/06/10% Sun Microsystems"

// Representation of predefined types:
//
// Instances of this class are inserted into the global context by
// the front end before the start of parsing.

/*
** DEPENDENCIES: ast_concrete_type.hh, utl_scoped_name.hh, utl_strlist.hh,
**		 ast_decl.hh
**
** USE: Included from ast.hh
*/

class	AST_PredefinedType : public virtual AST_ConcreteType
{
public:
  // Enum for all the different predefined types
  enum PredefinedType {
        PT_long		// Predefined type "long"
      , PT_ulong	// Predefined type "unsigned long"
      , PT_short	// Predefined type "short"
      , PT_ushort	// Predefined type "unsigned short"
      , PT_float	// Predefined type "float"
      , PT_double	// Predefined type "double"
      , PT_char		// Predefined type "char"
      , PT_boolean	// Predefined type "boolean"
      , PT_octet	// Predefined type "octet"
      , PT_any		// Predefined type "any"
      , PT_void		// Predefined type "void"
      , PT_pseudo	// Predefined type for pseudo objects
  };

  // Operations

  // Constructor(s)
  AST_PredefinedType();
  AST_PredefinedType(PredefinedType t, UTL_ScopedName *n, UTL_StrList *p);

  // Data Accessors
  const PredefinedType	pt();

  // Narrowing
  DEF_NARROW_METHODS1(AST_PredefinedType, AST_ConcreteType);
  DEF_NARROW_FROM_DECL(AST_PredefinedType);

  // AST Dumping
  virtual void		dump(ostream &o);

private:
  // Data
  const PredefinedType	pd_pt;		// The specific predefined type
};  

#endif           // _AST_PREDEFINED_TYPE_AST_PREDEFINED_TYPE_HH
